# DOC
#
# testcov.tcl - simple Tcl package for recording the test coverage
#
# Copyright (C) 2001 Arjen Markus
#
#
# General information:
#
# TESTCOV is a simple Tcl package that instruments Tcl source files,
# so that the statement coverage and basic branch coverage can be
# recorded and inspected.
# To do so, it uses the AWKPROC facility to analyse the source code
# and create an instrumented version. It uses the same facility to
# report the coverage.
#
# This package contains the following procedures:
# __cov_proc:    Register the use of a procedure
# __cov_branch:  Register the use of a branching construct
# __cov_loop:    Register the use of a looping construct
# __cov_save:    Save the registered information
# exit:          Overloaded version of the exit command
# source:        Overloaded version of the source command
#
# Usage:
# "testcov.tcl -inst source-file(s)" instruments the source code
# "testcov.tcl -rep source-file(s)" reports the coverage
# "source testcov.tcl" to incorporate the procedures
#
# version 0.3: corrected proc exit and messages about items not
#              covered yet, july 2001
#              (The mistake in "exit" was pointed out by Bernhard
#              Wallner, Germany)
# version 0.2: improved detail in coverage, june 2001
# version 0.1: initial implementation, june 2001
#
# ENDDOC
#
#  $Author$
#  $Date$
#  $Source$
#  $Log$
#

# --------------------------------------------------------------------
# Global procedures
# --------------------------------------------------------------------

# --------------------------------------------------------------------
#   Procedure: __cov_proc
#   Author:    Arjen Markus
#   Purpose:   Register the use of a procedure
#   Context:   Used in the instrumented source
#   Summary:
#              Increase the associated array element.
#   Arguments:
#   filename   Name of the source file
#   lineno     Line number in the source file
#   noargs     Number of extra arguments (treat the args argument)
# --------------------------------------------------------------------
#
proc __cov_proc { filename lineno noargs } {
   global __cov_state

   set noargs_reg $noargs
   if { $noargs >= 2 } { set noargs_reg 2 }

   if { [info exists __cov_state($filename,$lineno,$noargs_reg)] } {
      incr __cov_state($filename,$lineno,$noargs_reg)
   } else {
      set __cov_state($filename,$lineno,$noargs_reg) 1
   }

   if { [info exists __cov_state($filename,$lineno)] } {
      incr __cov_state($filename,$lineno)
   } else {
      set __cov_state($filename,$lineno) 1
   }

   return
}

# --------------------------------------------------------------------
#   Procedure: __cov_branch
#   Author:    Arjen Markus
#   Purpose:   Register the use of branching constructs
#   Context:   Used in the instrumented source
#   Summary:
#              Increase the associated array elements:
#              code = 0: before the branching
#              code = 1: the "true" branch
#              code > 1: alternative branches
#   Arguments:
#   filename   Name of the source file
#   lineno     Line number in the source file
#   code       Which code (or branch)?
# --------------------------------------------------------------------
#
proc __cov_branch { filename lineno code } {
   global __cov_state

   if { [info exists __cov_state($filename,$lineno,$code)] } {
      incr __cov_state($filename,$lineno,$code)
   } else {
      set __cov_state($filename,$lineno,$code) 1
   }

   return
}

# --------------------------------------------------------------------
#   Procedure: __cov_loop
#   Author:    Arjen Markus
#   Purpose:   Register the use of looping constructs
#   Context:   Used in the instrumented source
#   Summary:
#              Increase the associated array elements:
#              code = 0: before the loop
#              code = 1: any iteration
#   Arguments:
#   filename   Name of the source file
#   lineno     Line number in the source file
#   code       Which code?
# --------------------------------------------------------------------
#
proc __cov_loop { filename lineno code } {
   global __cov_state

   if { [info exists __cov_state($filename,$lineno,$code)] } {
      incr __cov_state($filename,$lineno,$code)
   } else {
      set __cov_state($filename,$lineno,$code) 1
      set __cov_state($filename,$lineno,count) 0
   }

   #
   # Special care for registering zero-iteration loops
   #
   if { $code == 0 } {
      set __cov_state($filename,$lineno,start) 1
   } else {
      if { $__cov_state($filename,$lineno,start) == 1 } {
         incr __cov_state($filename,$lineno,count)
         set  __cov_state($filename,$lineno,start) 0
      }
   }

   return
}

# --------------------------------------------------------------------
#   Procedure: __cov_save
#   Author:    Arjen Markus
#   Purpose:   Save the state
#   Context:   Used in the instrumented source
#   Summary:
#              Write the state array to the file "testcov.sav"
#   Arguments:
#              -- none --
# --------------------------------------------------------------------
#
proc __cov_save { } {
   global __cov_state

   set outfile [ open "testcov.sav" "w" ]
   puts $outfile "array set __cov_state [list [array get __cov_state]]"
   close $outfile

   return
}

# --------------------------------------------------------------------
#   Procedure: exit (aliased), source (aliased)
#   Author:    Arjen Markus
#   Purpose:   Save the state and exit or source an instrumented version
#   Context:   Used in the instrumented source
#   Summary:
#              Save the state array and then exit
#   Arguments:
#              -- none --
# --------------------------------------------------------------------
#
if { [info commands __cov_exit] == {} } {
   rename exit   __cov_exit
   rename source __cov_source

   proc exit { {exitcode 0} } {
      __cov_save
      __cov_exit $exitcode
   }

   proc source { filename } {
      if { [file exists "${filename}_inst"] } {
         set mtime_org  [file mtime  "$filename"]
         set mtime_inst [file mtime  "${filename}_inst"]
         if { $mtime_org < $mtime_inst } {
            uplevel [list __cov_source "${filename}_inst" ]
         } else {
            uplevel [list __cov_source "$filename" ]
         }
      } else {
         uplevel [list __cov_source "$filename" ]
      }
   }
}

# --------------------------------------------------------------------
# Initialisation of the state array
# --------------------------------------------------------------------
#
global __cov_state
set __cov_state(init) 1

# --------------------------------------------------------------------
#   Main program:
#   Instrument or report
# --------------------------------------------------------------------
#
global argv0
if { [info script] == "$argv0" } {
   set dirname [file dirname [info script]]
   source [file join $dirname "awkproc.tcl"]

   global argv

#
# Instrumentation of the source file(s)
#
   if { [lindex $argv 0] == "-inst" } {
      #package require AwkProc

      BeginFile {
         global outfile
         set outfile [ open "${FILENAME}_inst" "w" ]
         puts $outfile {
if { ! [info exists __cov_state] } {
   source "testcov.tcl"
}}
      }

      EndFile {
         global outfile
         close $outfile
      }

      DefaultAction {
         global $outfile
         puts $outfile $LINE
      }

      RegPattern {^[ ]*proc.*\{$} {
         global outfile
         puts $outfile $LINE
         if { [regexp {[\{ ]args[\} ]} $LINE] } {
            puts $outfile "__cov_proc $FILENAME $NL \[llength \$args\]"
         } else {
            puts $outfile "__cov_proc $FILENAME $NL -1"
         }
         set LINE ""
      }

      RegPattern {^[ ]*if .*\{$} {
         global outfile
         puts $outfile "__cov_branch $FILENAME $NL 0"
         puts $outfile $LINE
         puts $outfile "__cov_branch $FILENAME $NL 1"
         set LINE ""
      }

      RegPattern {^[ ]*for .*\{$} {
         global outfile
         puts $outfile "__cov_loop $FILENAME $NL 0"
         puts $outfile $LINE
         puts $outfile "__cov_loop $FILENAME $NL 1"
         set LINE ""
      }

      RegPattern {^[ ]*foreach .*\{$} {
         global outfile
         puts $outfile "__cov_loop $FILENAME $NL 0"
         puts $outfile $LINE
         puts $outfile "__cov_loop $FILENAME $NL 1"
         set LINE ""
      }

      RegPattern {^[ ]*while .*\{$} {
         global outfile
         puts $outfile "__cov_loop $FILENAME $NL 0"
         puts $outfile $LINE
         puts $outfile "__cov_loop $FILENAME $NL 1"
         set LINE ""
      }

      ProcessFiles [lrange $argv 1 end]
  }

#
# Reporting the test coverage
#
   if { [lindex $argv 0] == "-rep" } {
      global outfile
      global __cov_state
      source "testcov.sav"

      set outfile [ open "testcov.report" "w" ]
      puts $outfile "Test coverage report"
      puts $outfile "--------------------\n"

      BeginFile {
         global outfile
         puts $outfile "File: $FILENAME\n"
      }

      DefaultAction {
         global outfile
         puts $outfile "      $LINE"
      }

      RegPattern {^[ ]*proc.*\{$} {
         global outfile
         global __cov_state
         if { [info exists __cov_state($FILENAME,$NL) ] } {
            puts $outfile [ format "%4d: %s" $__cov_state($FILENAME,$NL) $LINE ]
            if { ![info exists __cov_state($FILENAME,$NL,-1) ] } {
               if { ![info exists __cov_state($FILENAME,$NL,0) ] } {
                  puts $outfile "Note: Procedure never called with zero extra arguments"
               }
               if { ![info exists __cov_state($FILENAME,$NL,1) ] } {
                  puts $outfile "Note: Procedure never called with one extra argument"
               }
               if { ![info exists __cov_state($FILENAME,$NL,2) ] } {
                  puts $outfile "Note: Procedure never called with two or more extra arguments"
               }
            }
         } else {
            puts $outfile "===>  $LINE"
            puts $outfile "Note: Procedure never called"
         }
         set LINE ""
      }

      RegPattern {^[ ]*if .*\{$} {
         global outfile
         global __cov_state
         if { [info exists __cov_state($FILENAME,$NL,0) ] } {
            if { [info exists __cov_state($FILENAME,$NL,1) ] } {
               puts $outfile [ format "%4d: %s" $__cov_state($FILENAME,$NL,1) $LINE ]
               if { $__cov_state($FILENAME,$NL,0) < $__cov_state($FILENAME,$NL,0) } {
                  puts $outfile \
                     "ERROR: Impossible case - programming error in testcov!"
               } elseif { $__cov_state($FILENAME,$NL,0) == $__cov_state($FILENAME,$NL,1) } {
                  puts $outfile \
                     "Note: Alternative branches not executed"
               }
            } else {
               puts $outfile [ format "%4d: %s" $__cov_state($FILENAME,$NL,0) $LINE ]
               puts $outfile \
                  "Note: Primary branch not executed"
            }
         } else {
            puts $outfile "===>  $LINE"
            puts $outfile "Note: Block never executed"
         }
         set LINE ""
      }

      Pattern { elseif } {
         global outfile
         puts $outfile " ???  $LINE"
         puts $outfile "Note: No coverage information available"
         set LINE ""
      }

      Pattern {switch } {
         global outfile
         puts $outfile " ???  $LINE"
         puts $outfile "Note: No coverage information available"
         set LINE ""
      }

      RegPattern {^[ ]*for .*\{$} {
         global outfile
         global __cov_state
         if { [info exists __cov_state($FILENAME,$NL,0) ] &&
              [info exists __cov_state($FILENAME,$NL,1) ] } {
            puts $outfile [ format "%4d: %s" $__cov_state($FILENAME,$NL,1) $LINE ]
            if { $__cov_state($FILENAME,$NL,0) < $__cov_state($FILENAME,$NL,count) } {
               puts $outfile \
                  "ERROR: Impossible case - programming error in testcov!"
            } elseif { $__cov_state($FILENAME,$NL,0) == $__cov_state($FILENAME,$NL,count) } {
               puts $outfile \
                  "Note: No zero-iterations"
            }
         } else {
            puts $outfile "===>  $LINE"
            puts $outfile "Note: Block never executed"
         }
         set LINE ""
      }

      RegPattern {^[ ]*foreach .*\{$} {
         global outfile
         global __cov_state
         if { [info exists __cov_state($FILENAME,$NL,0) ] &&
              [info exists __cov_state($FILENAME,$NL,1) ] } {
            puts $outfile [ format "%4d: %s" $__cov_state($FILENAME,$NL,1) $LINE ]
            if { $__cov_state($FILENAME,$NL,0) < $__cov_state($FILENAME,$NL,count) } {
               puts $outfile \
                  "ERROR: Impossible case - programming error in testcov!"
            } elseif { $__cov_state($FILENAME,$NL,0) == $__cov_state($FILENAME,$NL,count) } {
               puts $outfile \
                  "Note: No zero-iterations"
            }
         } else {
            puts $outfile "===>  $LINE"
            puts $outfile "Note: Block never executed"
         }
         set LINE ""
      }

      RegPattern {^[ ]*while .*\{$} {
         global outfile
         global __cov_state
         if { [info exists __cov_state($FILENAME,$NL,0) ] &&
              [info exists __cov_state($FILENAME,$NL,1) ] } {
            puts $outfile [ format "%5d:%s" $__cov_state($FILENAME,$NL,1) $LINE ]
            if { $__cov_state($FILENAME,$NL,0) < $__cov_state($FILENAME,$NL,count) } {
               puts $outfile \
                  "ERROR: Impossible case - programming error in testcov!"
            } elseif { $__cov_state($FILENAME,$NL,0) == $__cov_state($FILENAME,$NL,count) } {
               puts $outfile \
                  "Note: No zero-iterations"
            }
         } else {
            puts $outfile "===>  $LINE"
            puts $outfile "Note: Block never executed"
         }
         set LINE ""
      }

      eval ProcessFiles [lrange $argv 1 end]

      close $outfile
   }
}
# end of main
